# Copyright 2025 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from pylegend._typing import (
    PyLegendSequence,
    PyLegendUnion,
    PyLegendList,
    PyLegendCallable,
    PyLegendMapping,
    PyLegendHashable,
)
from pylegend.core.language.shared.primitives.primitive import PyLegendPrimitiveOrPythonPrimitive

__all__: PyLegendSequence[str] = [
    "PyLegendAggFunc",
    "PyLegendAggList",
    "PyLegendAggDict",
    "PyLegendAggInput",
]


PyLegendAggFunc = PyLegendUnion[  # type: ignore[explicit-any]
    PyLegendCallable[..., PyLegendPrimitiveOrPythonPrimitive],
    str,
    np.ufunc,
]

PyLegendAggList = PyLegendList[PyLegendAggFunc]

PyLegendAggDict = PyLegendMapping[
    PyLegendHashable,
    PyLegendUnion[
        PyLegendAggFunc,
        PyLegendAggList
    ]
]

PyLegendAggInput = PyLegendUnion[
    PyLegendAggFunc,
    PyLegendAggList,
    PyLegendAggDict,
]
